import json as _json

def _llm_response(obj: any) -> str:
    """
    Extracts the top result from the LLM output
    """

    try:
        return obj["choices"][0]["message"]["content"]
    except:
        return None

def _llm_json(obj: any):
    """
    Extracts the top result from the LLM output
    and converts it to JSON
    """

    try:
        result = obj["choices"][0]["message"]["content"]
        return _json.loads(result)
    except:
        return None



import json as _json
import requests as _requests
from enum import Enum as _Enum


class OperationType(str, _Enum):
    POST = "POST"
    GET = "GET"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


class Operation():

    def __init__(self, id: str, type: OperationType, name: str, description: str,
                 url: str, path: str, requires_auth: bool, schema: any):
        """
        Holds the properties on an operation prepared for execution
        - id: The identifier of the operation
        - type: The type of operation
        - name: The name of the operation
        - description: The description of the operation
        - url: The url of the operation
        - path: The path to the operation
        - requires_auth: If the operation requires authentication
        - schema: The schema of the operation
        """

        self.id = id
        self.type = type
        self.name = name
        self.description = description
        self.url = url
        self.path = path
        self.requires_auth = requires_auth
        self.schema = schema

    def __repr__(self) -> str:
        return _json.dumps(self.__dict__)

    @classmethod
    def from_obj(self, obj):
        """
        Returns an instance of an operation from a dict object
        """

        return Operation(obj['id'], obj['type'], obj['name'], obj['description'],
                         obj['url'], obj['path'], obj['auth'], obj['schema'])

    def metadata(self):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "type": self.type,
            "url": self.url,
            "path": self.path,
            "auth": self.requires_auth,
            "schema": self.schema
        }

    def embedding_obj(self):
        return "; ".join([
            f"Name: {self.name}",
            f"Description: {self.description}",
            f"Type: {self.type}",
            f"URL: {self.url}",
            f"Path: {self.path}",
            f"Requires Auth: {self.requires_auth}",
            f"Schema: {self.schema}"
        ])

    def endpoint(self) -> str:
        return self.url + self.path

    def execute(self, params, body, auth_token=None):
        """
        Executes the command.

        Returns: The response provided by the execution API
        """

        result = None
        data = _json.dumps(body).encode('utf-8')

        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        if self.requires_auth:
            headers["Authorization"] = f"Bearer {auth_token}"

        if self.type == OperationType.POST:
            result = _requests.post(
                self.endpoint(),
                headers=headers,
                params=params,
                data=data
            )

        elif self.type == OperationType.GET:
            result = _requests.get(
                self.endpoint(),
                headers=headers,
                params=params,
                data=data
            )

        elif self.type == OperationType.PUT:
            result = _requests.put(
                self.endpoint(),
                headers=headers,
                params=params,
                data=data
            )

        elif self.type == OperationType.PATCH:
            result = _requests.patch(
                self.endpoint(),
                headers=headers,
                params=params,
                data=data
            )

        elif self.type == OperationType.DELETE:
            result = _requests.delete(
                self.endpoint(),
                headers=headers,
                params=params,
                data=data
            )

        if not result:
            return None

        return result.json()



import os as _os
import pinecone as _pinecone
import openai as _openai

from .operation_utils import *
from .operation import *
from .operator import *

def init(openai_key, pinecone_key, pinecone_region, pinecone_index):
    _openai.api_key = openai_key

    _pinecone.init(
        api_key=pinecone_key,
        environment=pinecone_region
    )

    _os.environ["PINECONE_INDEX"] = pinecone_index



import os as _os
import json as _json
from uuid import uuid4 as _uuid4
import pinecone as _pinecone
from openai.embeddings_utils import get_embedding as _get_embedding
from .operation import Operation as _Operation


class Utils():
    """
    A set of utility objects for managing operations
    """

    @classmethod
    def create_operation(self, namespace, type, name, description,
                         url, path, requires_auth, schema):
        """
        Creates an operation, creates an embedding from it, and
        stores it in a vector database.
        - namespace: The namespace to store the embedding
        - type: The type of operation
        - url: The url of the operation
        - path: The path to the operation
        - requires_auth: If the operation requires authentication
        - schema: The schema of the operation

        Also writes the operation ID to the `ops_list.txt` file

        Returns: The created operation
        """

        index = _pinecone.Index(_os.getenv("PINECONE_INDEX"))

        id = str(_uuid4())

        operation = _Operation(
            id=id,
            type=type,
            name=name,
            description=description,
            url=url,
            path=path,
            requires_auth=requires_auth,
            schema=_json.dumps(_json.loads(schema), separators=(',', ': '))
        )

        embedding = _get_embedding(operation.embedding_obj(),
                                  engine="text-embedding-ada-002")

        to_upsert = zip([id], [embedding], [operation.metadata()])

        index.upsert(vectors=list(to_upsert), namespace=namespace)

        return operation

    @classmethod
    def get_operation(self, namespace: str, id: str):
        """
        Fetches an existing operation from the vector database.
        - namespace: The namespace to store the embedding
        - id: The identifier of the operation

        Returns: The operation represented
        """
        index = _pinecone.Index(_os.getenv("PINECONE_INDEX"))

        result = index.fetch([id], namespace=namespace)
        vectors = result.get('vectors')
        vector = vectors.get(id)

        if not vector:
            return None

        obj = vector.get('metadata')
        return _Operation.from_obj(obj)

    @classmethod
    def update_operation(self, namespace, id, type, name, description,
                         url, path, requires_auth, schema):
        """
        Updates an existing operation, creates a new embedding, and
        overrides the existing operation in the vector database.
        - namespace: The namespace to store the embedding
        - id: The identifier of the operation
        - type: The type of operation
        - url: The url of the operation
        - path: The path to the operation
        - requires_auth: If the operation requires authentication
        - schema: The schema of the operation

        Returns: The updated operation
        """

        operation = _Operation(
            id=id,
            type=type,
            name=name,
            description=description,
            url=url,
            path=path,
            requires_auth=requires_auth,
            schema=_json.dumps(_json.loads(schema), separators=(',', ': '))
        )

        index = _pinecone.Index(_os.getenv("PINECONE_INDEX"))

        embedding = _get_embedding(operation.embedding_obj(),
                                  engine="text-embedding-ada-002")

        to_upsert = zip([id], [embedding], [operation.metadata()])

        index.upsert(vectors=list(to_upsert), namespace=namespace)

        return operation

    @classmethod
    def remove_operation(self, namespace: str, id: str):
        """
        Removes an existing operation from the vector database.
        - namespace: The namespace to store the embedding
        - id: The identifier of the operation

        Also removes the operation ID from `ops_list.txt` file.
        """

        index = _pinecone.Index(_os.getenv("PINECONE_INDEX"))

        index.delete(ids=[id], namespace=namespace)

    @classmethod
    def remove_namespace(self, namespace: str):
        """
        [DANGEROUS] Deletes an entire namespace of operations.
        - namespace: The namespace in the vector database
        """
        index = _pinecone.Index(_os.getenv("PINECONE_INDEX"))

        index.delete(deleteAll='true', namespace=namespace)



import os
import json
import pinecone
from openai.embeddings_utils import get_embedding
from openai import ChatCompletion
from .operation import Operation
from .operation_utils import Utils
from .utils import _llm_json, _llm_response


class Operator():

    def __init__(self, namespace: str) -> None:
        self.namespace = namespace

    def get(self, id: str) -> Operation:
        """
        Fetches a pre-determined operation
        - id: The operation identifier

        Returns: Operation
        """

        op = Utils.get_operation(self.namespace, id)
        if not op:
            raise ValueError("Operation does not exist")

        return Operation.from_obj(op)

    def find(self, prompt: str, top_k: int=3) -> list[Operation]:
        """
        Finds a set operations based on a provided prompt
        - prompt: The prompt to use for the search

        Returns set[Operation]
        """

        index = pinecone.Index(os.getenv("PINECONE_INDEX"))
        embedding = get_embedding(prompt, engine="text-embedding-ada-002")

        result = index.query(
            vector=embedding,
            top_k=top_k,
            namespace=self.namespace,
            include_metadata=True
        )

        operations = []
        for match in result.get('matches'):
            operations.append(Operation.from_obj(match.get('metadata')))

        return operations

    def pick(self, prompt: str, operations: list[Operation]) -> list[Operation]:
        """
        Given a prompt and a list of operations, the LLM selects
        the operation that best fits the prompt.
        - prompt: The prompt to base the selection off of
        - operations: The list of operations to choose from.

        Returns: The identifier of the operation.
        """

        clean_ops = [op.__dict__ for op in operations]
        response = ChatCompletion.create(
            model="gpt-4",
            messages=[
                {"role": "system", "content": """
                Given a list of operations, pick one that would best contribute to the user's prompt.
                """.replace("\n", " ")},
                {"role": "system", "content": "Output the ID of the operation."},
                {"role": "user", "content": f"Operations: {json.dumps(clean_ops)}"},
                {"role": "user", "content": f"Prompt: {prompt}"},
                {"role": "user", "content": "Output the ID of the operation and nothing more."}
            ],
            temperature=0.0
        )

        op_id = _llm_response(response)
        operation = None
        for op in operations:
            if op.id == op_id:
                operation = op
                break

        return operation

    def prepare(self, prompt: str, operation: Operation):
        """
        Generates a payload for the operation
        based on the prompt and operation schema.
        - prompt: The prompt to use
        - operation: The operation to prepare for

        Returns: JSON object with params and body
        """

        response = ChatCompletion.create(
            model="gpt-4",
            messages=[
                {"role": "system", "content": """
                Give an operation with a predefined schema and a user prompt,
                provide parameter and body values to send to the operation based on the prompt.
                """.replace("\n", " ")},
                {"role": "system", "content": "Output in JSON format"},
                {"role": "user", "content": f"Operation: {operation.__dict__}"},
                {"role": "user", "content": f"Prompt: {prompt}"},
                {"role": "user", "content": "Output the params and body in JSON format and nothing more."}
            ],
            temperature=0.0
        )

        return _llm_json(response)

    def execute(self, operation: Operation, params: any, body: any, auth_token: str=None):
        """
        Executes the provided operation.

        Returns: Value from operation
        """

        return operation.execute(params=params, body=body, auth_token=auth_token)

    def react(self, prompt: str, operation: Operation, values: str, result: str) -> str:
        """
        Reacts to the operation execution based on
        the original prompt

        Returns: The LLM reaction response
        """
        response = ChatCompletion.create(
            model="gpt-4",
            messages=[
                {"role": "system", "content": """
                Given a the original prompt and the execution result of an operation that followed,
                respond to the prompt based on the execution result.
                """.replace("\n", " ")},
                {"role": "user", "content": f"Prompt: {prompt}"},
                {"role": "user", "content": f"Operation: {operation.__dict__}"},
                {"role": "user", "content": f"Values passed to operation: {values}"},
                {"role": "user", "content": f"Execution result: {result}"}
            ],
            temperature=0.0
        )

        return _llm_response(response)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gptop",
    version="0.0.5",
    author="Nick Crews",
    description="Handles communication with GPTs",
    keywords="gptop GPT operator llm ai",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=[
        "tests",
        "scripts",
        "example",
        "cli"
    ]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=[
        "requests >= 2.20",
        "scikit-learn >= 1.0.2",
        "tenacity >= 8.0.1",
        "matplotlib",
        "plotly",
        "numpy",
        "scipy",
        "pandas >= 1.2.3",
        "pandas-stubs >= 1.1.0.11",
        "openpyxl >= 3.0.7",
        "aiohttp",
        "tqdm",
        "pyyaml >= 5.4",
        "loguru >= 0.5.0",
        "typing-extensions >= 3.7.4",
        "dnspython >= 2.0.0",
        "python_dateutil >= 2.5.3",
        "urllib3 >= 1.21.1",
        "openai",
        "pinecone-client",
    ]
)
